@@ -9,26 +9,24 @@
 
 use std::collections::HashMap;
 
-#[doc(hidden)]
-pub mod wapc;
+use vino_wapc::{exports::*, OutputSignal};
 
-use vino_codec::messagepack::{
-  deserialize,
-  serialize,
-};
+/// Module that encapsulates a linked provider as input.
+mod provider_link;
+
+/// Module for log-like functionality in WASM.
+pub mod log;
+
+use vino_codec::messagepack::{deserialize, serialize};
 
 /// Errors for WebAssembly providers.
 pub mod error;
 /// The WebAssembly implementation of a Port Sender.
-pub mod port_sender;
+mod port_sender;
 pub use error::Error;
 pub use port_sender::PortSender;
-use vino_packet::{
-  v0,
-  Packet,
-};
+use vino_packet::{v0, Packet};
 
-use crate::wasm::wapc::*;
 type Result<T> = std::result::Result<T, Error>;
 
 /// The return signature for WebAssembly jobs.
@@ -40,55 +38,64 @@ pub type CallResult = Result<Vec<u8>>;
 
 /// Common imports for WebAssembly providers and components.
 pub mod prelude {
+  pub use vino_transport::error::TransportError;
+  pub use vino_transport::{MessageTransport, TransportMap, TransportWrapper};
+  pub use vino_types::*;
+
+  pub use super::error::ComponentError;
+  pub use super::provider_link::{PortOutput, ProviderOutput, WasmProviderLink};
   pub use super::{
-    console_log,
-    wapc,
-    CallResult,
-    Dispatch,
-    Error,
-    IncomingPayload,
-    JobResult,
-    PortSender,
+    console_log, CallResult, Dispatch, Error as WasmError, IncomingPayload, JobResult, PortSender,
     WapcComponent,
   };
-  pub use crate::codec::messagepack::{
-    deserialize,
-    serialize,
-  };
+  pub use crate::codec::messagepack::{deserialize, serialize};
+  pub use crate::provider_link::ProviderLink;
+  pub use crate::wasm::log;
+  pub use vino_wapc::exports as wapc;
+  pub use vino_wapc::*;
 }
 
-use crate::OutputSignal;
+fn serialize_payload(id: u32, packet: Option<v0::Payload>) -> Result<Vec<u8>> {
+  let bytes = match packet {
+    Some(packet) => {
+      let bytes = serialize(&Packet::V0(packet))?;
+      let mut payload = Vec::with_capacity(bytes.len() + 4);
+      payload.extend_from_slice(&id.to_be_bytes());
+      payload.extend(bytes.into_iter());
+      payload
+    }
+    None => {
+      let mut payload = Vec::with_capacity(4);
+      payload.extend_from_slice(&id.to_be_bytes());
+      payload
+    }
+  };
+  Ok(bytes)
+}
 
 /// Send a [Packet] out the named port.
-pub fn port_send(port_name: &str, packet: v0::Payload) -> Result<()> {
-  host_call(
-    "",
-    port_name,
-    OutputSignal::Output.as_str(),
-    &serialize(&Packet::V0(packet))?,
-  )?;
+pub fn port_send(port_name: &str, id: u32, packet: v0::Payload) -> Result<()> {
+  let bytes = serialize_payload(id, Some(packet))?;
+  host_call("0", port_name, OutputSignal::Output.as_str(), &bytes)?;
   Ok(())
 }
 
 /// Send a [Packet] out the named port and immediately close it.
-pub fn port_send_close(port_name: &str, packet: v0::Payload) -> Result<()> {
-  host_call(
-    "",
-    port_name,
-    OutputSignal::OutputDone.as_str(),
-    &serialize(&Packet::V0(packet))?,
-  )?;
+pub fn port_send_close(port_name: &str, id: u32, packet: v0::Payload) -> Result<()> {
+  let bytes = serialize_payload(id, Some(packet))?;
+  host_call("0", port_name, OutputSignal::OutputDone.as_str(), &bytes)?;
   Ok(())
 }
 
 /// Close the referenced port.
-pub fn port_close(port_name: &str) -> Result<()> {
-  host_call("", port_name, OutputSignal::Done.as_str(), &[])?;
+pub fn port_close(port_name: &str, id: u32) -> Result<()> {
+  let bytes = serialize_payload(id, None)?;
+  host_call("0", port_name, OutputSignal::Done.as_str(), &bytes)?;
   Ok(())
 }
 
 /// The function through which all host calls take place.
-pub(crate) fn host_call(binding: &str, ns: &str, op: &str, msg: &[u8]) -> CallResult {
+pub fn host_call(binding: &str, ns: &str, op: &str, msg: &[u8]) -> CallResult {
   let callresult = unsafe {
     __host_call(
       binding.as_ptr(),
@@ -101,16 +108,19 @@ pub(crate) fn host_call(binding: &str, ns: &str, op: &str, msg: &[u8]) -> CallRe
       msg.len(),
     )
   };
+
   if callresult != 1 {
     // call was not successful
-    let errlen = unsafe { __host_error_len() };
-    let buf = Vec::with_capacity(errlen);
+    let len = unsafe { __host_error_len() };
+    let buf = Vec::with_capacity(len);
     let retptr = buf.as_ptr();
     let _slice = unsafe {
       __host_error(retptr);
-      std::slice::from_raw_parts(retptr, errlen)
+      std::slice::from_raw_parts(retptr, len)
     };
-    Err(Error::HostError("Component execution failed".to_owned()))
+    Err(Error::HostError(
+      String::from_utf8_lossy(_slice).to_string(),
+    ))
   } else {
     // call succeeded
     let len = unsafe { __host_response_len() };
@@ -135,18 +145,27 @@ pub fn console_log(s: &str) {
 
 /// A map of port name to payload message.
 pub struct IncomingPayload {
+  id: u32,
   encoded: HashMap<String, Vec<u8>>,
 }
 
 impl IncomingPayload {
   /// Decode MessagePack bytes into an [IncomingPayload].
   pub fn from_buffer(buffer: &[u8]) -> Result<Self> {
-    let input_encoded: HashMap<String, Vec<u8>> = deserialize(buffer)?;
+    let (id, input_encoded): (u32, HashMap<String, Vec<u8>>) = deserialize(buffer)?;
 
     Ok(Self {
+      id,
       encoded: input_encoded,
     })
   }
+
+  /// Get the transaction ID associated with this [IncomingPayload].
+  #[must_use]
+  pub fn id(&self) -> u32 {
+    self.id
+  }
+
   /// Get the contained bytes for the specified port.
   pub fn get(&self, port: &str) -> Result<&Vec<u8>> {
     self
